#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['zwi', 'zwi.scripts']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

entry_points = \
{'console_scripts': ['zwi = zwi.scripts.zwi:main',
                     'zwibok = zwi.scripts.zwibok:main']}

setup(name='zwi',
      version='0.3a2.dev1',
      description='Zwift data explorations.',
      author=None,
      author_email='Damon Permezel <permezel@me.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      entry_points=entry_points,
      python_requires='>=3.9',
     )
